"""__init__.py."""
from importlib_metadata import version

from konoha.sentence_tokenizer import SentenceTokenizer  # NOQA
from konoha.word_tokenizer import WordTokenizer  # NOQA

__version__ = version("konoha")



from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class CharacterTokenizer(BaseTokenizer):
    def __init__(self):
        super().__init__(name="character")

    def tokenize(_, text: str):
        return [Token(surface=char) for char in list(text)]



from typing import List

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class WhitespaceTokenizer(BaseTokenizer):
    """Simple rule-based word tokenizer."""

    def __init__(_) -> None:
        super().__init__(name="whitespace")

    def tokenize(_, text: str) -> List[Token]:
        return [Token(surface=surface) for surface in text.split(" ")]



from .character_tokenizer import CharacterTokenizer  # NOQA
from .janome_tokenizer import JanomeTokenizer  # NOQA
from .kytea_tokenizer import KyTeaTokenizer  # NOQA
from .mecab_tokenizer import MeCabTokenizer  # NOQA
from .nagisa_tokenizer import NagisaTokenizer  # NOQA
from .sentencepiece_tokenizer import SentencepieceTokenizer  # NOQA
from .sudachi_tokenizer import SudachiTokenizer  # NOQA
from .whitespace_tokenizer import WhitespaceTokenizer  # NOQA



from typing import List

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class NagisaTokenizer(BaseTokenizer):
    def __init__(self) -> None:
        from nagisa import Tagger
        super().__init__(name="nagisa")
        self._tokenizer = Tagger()

    def tokenize(self, text: str) -> List[Token]:
        response = self._tokenizer.tagging(text)
        tokens = [
            Token(surface=surface, postag=postag) for (surface, postag) in zip(response.words, response.postags)
        ]
        return tokens



from typing import List
from typing import Optional

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class JanomeTokenizer(BaseTokenizer):
    def __init__(self, user_dictionary_path: Optional[str] = None) -> None:
        from janome.tokenizer import Tokenizer
        super().__init__(name="janome")
        self._tokenizer = Tokenizer(udic=user_dictionary_path)

    def tokenize(self, text: str) -> List[Token]:
        return_result = []
        for morph in self._tokenizer.tokenize(text):
            postag, postag2, postag3, postag4 = morph.part_of_speech.split(",")
            token = Token(
                surface=morph.surface,
                postag=postag,
                postag2=postag2,
                postag3=postag3,
                postag4=postag4,
                inflection=morph.infl_type,
                conjugation=morph.infl_form,
                base_form=morph.base_form,
                yomi=morph.reading,
                pron=morph.phonetic,
            )
            return_result.append(token)

        return return_result



from typing import List

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class SentencepieceTokenizer(BaseTokenizer):
    def __init__(self, model_path: str) -> None:
        from sentencepiece import SentencePieceProcessor
        super().__init__(name="sentencepiece")
        self._tokenizer = SentencePieceProcessor()
        self._tokenizer.load(model_path)

    def tokenize(self, text: str) -> List[Token]:
        result = []
        for subword in self._tokenizer.EncodeAsPieces(text):
            token = Token(surface=subword)
            result.append(token)
        return result



from typing import List
from typing import Optional

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class KyTeaTokenizer(BaseTokenizer):
    def __init__(self, model_path: Optional[str] = None) -> None:
        from Mykytea import Mykytea
        super().__init__(name="kytea")

        kytea_option = ""
        if model_path is not None:
            kytea_option += "-model {}".format(model_path)
        self._tokenizer = Mykytea(kytea_option)

    def tokenize(self, text: str) -> List[Token]:
        tokens = []
        response = self._tokenizer.getTagsToString(text)
        # FIXME Following dirty workaround is required to
        #       process inputs which include <whitespace> itself
        #       (e.g. "私 は猫")
        response = response.replace("\\ ", "<SPACE>").replace("  ", " <SPACE>")
        for elem in response.split(" ")[:-1]:
            # FIXME If input contains a character "/",
            #       KyTea outputs "//補助記号/・",
            #       which breaks the simple logic elem.split("/")
            pron, postag, surface = map(lambda e: e[::-1], elem[::-1].split("/", maxsplit=2))
            surface = surface.replace("<SPACE>", " ")
            tokens.append(Token(surface=surface, postag=postag, pron=pron))

        return tokens



from abc import ABC
from abc import abstractmethod
from typing import List

from konoha.data.token import Token


class BaseTokenizer(ABC):
    """Base class for word levelkonoha.tokenizer"""

    def __init__(self, name: str) -> None:
        self._name = name

    @abstractmethod
    def tokenize(_, text: str) -> List[Token]:
        """Abstract method forkonoha.tokenization"""
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Return name ofkonoha.tokenizer"""
        return self._name



"""Token class."""
from typing import Dict
from typing import List
from typing import Optional


class Token:
    """Token class for konoha."""

    def __init__(
        self,
        surface: str,
        postag: Optional[str] = None,
        postag2: Optional[str] = None,
        postag3: Optional[str] = None,
        postag4: Optional[str] = None,
        inflection: Optional[str] = None,
        conjugation: Optional[str] = None,
        base_form: Optional[str] = None,
        yomi: Optional[str] = None,
        pron: Optional[str] = None,
        normalized_form: Optional[str] = None,
    ):
        """
        Initializer for Token.

        Parameters
        ---
        surface (str)
            surface (original form) of a word
        postag (str, default: None)
            part-of-speech tag of a word (optional)
        postag2 (str, default: None)
            detailed part-of-speech tag of a word (optional)
        postag3 (str, default: None)
            detailed part-of-speech tag of a word (optional)
        postag4 (str, default: None)
            detailed part-of-speech tag of a word (optional)
        inflection (str, default: None)
            conjugate type of word (optional)
        conjugation (str, default: None)
            conjugate type of word (optional)
        base_form (str, default: None)
            base form of a word
        yomi (str, default: None)
            yomi of a word (optional)
        pron (str, default: None)
            pronounciation of a word (optional)
        normalized_form (str, default: None)
            normalized form of a word (optional)
            Note that normalized_form is only
            available on SudachiPy
        """
        self._surface = surface
        self._postag = postag
        self._postag2 = postag2
        self._postag3 = postag3
        self._postag4 = postag4
        self._inflection = inflection
        self._conjugation = conjugation
        self._base_form = base_form
        self._yomi = yomi
        self._pron = pron
        self._normalized_form = normalized_form

    def __repr__(self) -> str:
        return self.surface

    def __eq__(self, right) -> bool:
        return self.surface == right.surface and self.feature == right.feature

    @property
    def surface(self):
        return self._surface

    @property
    def postag(self):
        return self._postag

    @property
    def postag2(self):
        return self._postag2

    @property
    def postag3(self):
        return self._postag3

    @property
    def postag4(self):
        return self._postag4

    @property
    def inflection(self):
        return self._inflection

    @property
    def conjugation(self):
        return self._conjugation

    @property
    def base_form(self):
        return self._base_form

    @property
    def yomi(self):
        return self._yomi

    @property
    def pron(self):
        return self._pron

    @property
    def normalized_form(self):
        return self._normalized_form

    @property
    def feature_names(_) -> List[str]:
        return [
            "surface",
            "postag",
            "postag2",
            "postag3",
            "postag4",
            "inflection",
            "conjugation",
            "base_form",
            "yomi",
            "pron",
            "normalized_form",
        ]

    @property
    def feature(self) -> str:
        feature = []
        if self.postag is not None:
            feature.append(self.postag)
        if self.postag2 is not None:
            feature.append(self.postag2)
        if self.postag3 is not None:
            feature.append(self.postag3)
        if self.postag4 is not None:
            feature.append(self.postag4)
        if self.inflection is not None:
            feature.append(self.inflection)
        if self.conjugation is not None:
            feature.append(self.conjugation)
        if self.base_form is not None:
            feature.append(self.base_form)
        if self.yomi is not None:
            feature.append(self.yomi)
        if self.pron is not None:
            feature.append(self.pron)
        if self.normalized_form is not None:
            feature.append(self.normalized_form)
        return ",".join(feature)

    @classmethod
    def from_dict(cls, dict: dict):
        return cls(**dict)

    def dict(self) -> Dict:
        return {f: getattr(self, f) for f in self.feature_names}



import logging
import os.path
import re
import secrets
from typing import Optional

RED = "\033[1;31m"
RST = "\033[0;0m"


logger = logging.getLogger(__name__)


class Resource:
    NBYTES = 6
    KONOHA_DIR = "/tmp/konoha"

    def __init__(self, path: Optional[str]) -> None:
        if path is None:
            self._path = None
            self._raw_path = None

        elif path.startswith("s3://"):
            self._path = self.download_from_s3(path)
            self._raw_path = path

        else:
            self._path = path
            self._raw_path = path

    def download_from_s3(self, path: str) -> str:
        """
        Download file(s) from Amazon S3.

        """
        try:
            import boto3
        except ImportError:
            msg = "Please install boto3:"
            msg += f" {RED}`pip install boto3`{RST}"
            msg += f" or {RED}`pip install konoha[remote]{RST}"
            raise ImportError(msg)

        prefix = secrets.token_hex(self.NBYTES)
        resource_dir = os.path.join(self.KONOHA_DIR, prefix)
        os.makedirs(resource_dir, exist_ok=True)

        re_result = re.search(r"s3://(.*?)/(.+)", path)
        if re_result is None:
            raise ValueError(f"Cannot find bucket name or prefix from {path}")

        bucket_name = re_result.group(1)
        prefix = re_result.group(2)

        resource = boto3.resource("s3")
        bucket = resource.Bucket(bucket_name)

        for obj in bucket.objects.filter(Prefix=prefix):
            if obj.size == 0:
                data_dir = os.path.join(resource_dir, prefix)
            else:
                data_dir = os.path.join(resource_dir, os.path.dirname(obj.key))
            os.makedirs(data_dir, exist_ok=True)

            logger.info(f"Downloading {obj.key}")
            download_path = os.path.join(resource_dir, obj.key)
            bucket.download_file(obj.key, download_path)
            logging.info(f"Downloaded to {download_path}")

        dest_dir = os.path.join(resource_dir, prefix)
        logging.info(f"Downloaded to {dest_dir}")
        return dest_dir

    @property
    def path(self) -> Optional[str]:
        return self._path


if __name__ == "__main__":
    r = Resource("/tmp/main")
    print(r.path)



from typing import List

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class SudachiTokenizer(BaseTokenizer):
    def __init__(self, mode: str) -> None:
        from sudachipy import dictionary
        from sudachipy import tokenizer
        super().__init__(name="sudachi ({})".format(mode))

        try:
            self._tokenizer = dictionary.Dictionary().create()
        except KeyError:
            msg = "Loading a dictionary fails."
            msg += " ( see https://github.com/WorksApplications/SudachiPy#install-dict-packages )"  # NOQA
            raise KeyError(msg)

        _mode = mode.capitalize()
        if _mode == "A":
            self._mode = tokenizer.Tokenizer.SplitMode.A
        elif _mode == "B":
            self._mode = tokenizer.Tokenizer.SplitMode.B
        elif _mode == "C":
            self._mode = tokenizer.Tokenizer.SplitMode.C
        else:
            raise ValueError("Invalid mode is specified. Mode should be A, B, or C.")  # NOQA

    def tokenize(self, text: str) -> List[Token]:
        result = []
        for token in self._tokenizer.tokenize(text, self._mode):
            (postag, postag2, postag3, postag4, inflection, conjugation) = token.part_of_speech()
            result.append(
                Token(
                    surface=token.surface(),
                    postag=postag,
                    postag2=postag2,
                    postag3=postag3,
                    postag4=postag4,
                    inflection=inflection,
                    conjugation=conjugation,
                    base_form=token.dictionary_form(),
                    normalized_form=token.normalized_form(),
                    yomi=token.reading_form(),
                )
            )
        return result



import logging

from fastapi import FastAPI
from rich.logging import RichHandler

from konoha.api.v1 import tokenization
from konoha.api.v1 import batch_tokenization


FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)


def create_app() -> FastAPI:
    app = FastAPI()
    app.state.cache = {}
    app.include_router(tokenization.router)
    app.include_router(batch_tokenization.router)
    return app



"""Word Level Tokenizer."""
import warnings
from typing import Dict
from typing import List
from typing import Optional

import requests

from konoha import word_tokenizers
from konoha.data.resource import Resource
from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


class WordTokenizer:
    def __init__(
        self,
        tokenizer: str = "MeCab",
        user_dictionary_path: Optional[str] = None,
        system_dictionary_path: Optional[str] = None,
        model_path: Optional[str] = None,
        mode: Optional[str] = None,
        dictionary_format: Optional[str] = None,
        endpoint: Optional[str] = None,
        ssl: Optional[bool] = None,
        port: Optional[int] = None,
    ) -> None:
        user_dictionary = Resource(user_dictionary_path)
        system_dictionary = Resource(system_dictionary_path)
        model = Resource(model_path)

        self._tokenizer_name = tokenizer.lower()
        self._user_dictionary_path = user_dictionary.path
        self._system_dictionary_path = system_dictionary.path
        self._model_path = model.path
        self._mode = mode.lower() if mode is not None else None
        self._dictionary_format = dictionary_format
        self._tokenizer: Optional[BaseTokenizer] = None
        self._endpoint = endpoint
        self._ssl = ssl
        self._port = port

        if not isinstance(endpoint, str):
            self._setup_tokenizer()

    def _setup_tokenizer(self) -> None:
        if self._tokenizer_name == "character":
            self._tokenizer = word_tokenizers.CharacterTokenizer()

        if self._tokenizer_name == "whitespace":
            self._tokenizer = word_tokenizers.WhitespaceTokenizer()

        if self._tokenizer_name == "kytea":
            self._tokenizer = word_tokenizers.KyTeaTokenizer(model_path=self._model_path)

        if self._tokenizer_name == "sentencepiece":
            if self._model_path is None:
                raise ValueError("`model_path` must be specified for sentencepiece.")

            self._tokenizer = word_tokenizers.SentencepieceTokenizer(model_path=self._model_path)

        if self._tokenizer_name == "mecab":
            self._tokenizer = word_tokenizers.MeCabTokenizer(
                user_dictionary_path=self._user_dictionary_path,
                system_dictionary_path=self._system_dictionary_path,
                dictionary_format=self._dictionary_format,
            )

        if self._tokenizer_name == "janome":
            self._tokenizer = word_tokenizers.JanomeTokenizer(user_dictionary_path=self._user_dictionary_path)

        if self._tokenizer_name == "sudachi":
            if self._mode is None:
                raise ValueError("`mode` must be specified for sudachi.")

            self._tokenizer = word_tokenizers.SudachiTokenizer(mode=self._mode)

        if self._tokenizer_name == "nagisa":
            self._tokenizer = word_tokenizers.NagisaTokenizer()

    def tokenize(self, text: str) -> List[Token]:
        """Tokenize input text"""

        if isinstance(self._endpoint, str):
            endpoint = self.get_endpoint("/api/v1/tokenize")
            payload = dict(self.payload, text=text)
            token_params = self._tokenize_with_remote_host(endpoint=endpoint, payload=payload, headers=self.headers)
            return [Token.from_dict(token_param) for token_param in token_params]

        else:
            assert self._tokenizer is not None
            return self._tokenizer.tokenize(text)

    def batch_tokenize(self, texts: List[str]) -> List[List[Token]]:
        """Tokenize input texts"""

        if isinstance(self._endpoint, str):
            endpoint = self.get_endpoint("/api/v1/batch_tokenize")
            payload = dict(self.payload, texts=texts)
            token_params_list = self._batch_tokenize_with_remote_host(
                endpoint=endpoint,
                payload=payload,
                headers=self.headers,
            )

            tokens_list: List[List[Token]] = []
            for tokens in token_params_list:
                tokens_list.append([Token.from_dict(token) for token in tokens])

            return tokens_list

        else:
            assert self._tokenizer is not None
            return [self._tokenizer.tokenize(text) for text in texts]

    @staticmethod
    def _tokenize_with_remote_host(endpoint: str, payload: Dict, headers: Dict) -> List[Dict]:
        return requests.post(endpoint, json=payload, headers=headers).json()["tokens"]

    @staticmethod
    def _batch_tokenize_with_remote_host(endpoint: str, payload: Dict, headers: Dict) -> List[List[Dict]]:
        return requests.post(endpoint, json=payload, headers=headers).json()["tokens_list"]

    @property
    def tokenizer(self) -> BaseTokenizer:
        warnings.warn("attribute `tokenizer` will be removed in v5.0.0.")
        assert self._tokenizer is not None
        return self._tokenizer

    @property
    def name(self) -> str:
        assert self._tokenizer is not None
        return self._tokenizer.name

    def get_endpoint(self, method: str):
        assert self._endpoint
        endpoint = self._endpoint

        if self._ssl and not endpoint.startswith("https://"):
            endpoint = "https://" + endpoint

        if not self._ssl and not endpoint.startswith("http://"):
            endpoint = "http://" + endpoint

        return endpoint + method

    @property
    def payload(self):
        return {
            "tokenizer": self._tokenizer_name,
            "user_dictionary_path": self._user_dictionary_path,
            "system_dictionary_path": self._system_dictionary_path,
            "model_path": self._model_path,
            "mode": self._mode,
            "dictionary_format": self._dictionary_format,
        }

    @property
    def headers(self):
        return {"Content-Type": "application/json"}



from typing import List
from typing import Optional

from konoha.data.token import Token
from konoha.word_tokenizers.tokenizer import BaseTokenizer


def parse_feature_for_ipadic(elem) -> Token:
    surface, feature = elem.split("\t")
    (postag, postag2, postag3, postag4, inflection, conjugation, base_form, *other) = feature.split(",")

    # For words not in a dictionary
    if len(other) == 2:
        yomi, pron = other
    else:
        yomi, pron = None, None

    return Token(
        surface=surface,
        postag=postag,
        postag2=postag2,
        postag3=postag3,
        postag4=postag4,
        inflection=inflection,
        conjugation=conjugation,
        base_form=base_form,
        yomi=yomi,
        pron=pron,
    )


def parse_feature_for_unidic(elem) -> Token:
    """
    UniDic format: https://unidic.ninjal.ac.jp/faq
    """
    surface, feaature_line = elem.split("\t")
    features = feaature_line.split(",")

    postag = features[0] or None
    postag2 = features[1] or None
    postag3 = features[2] or None
    postag4 = features[3] or None
    inflection = features[4] or None
    conjugation = features[5] or None

    if len(features) >= 10:
        yomi = features[6] or None
        base_form = features[7] or None
        pron = features[9] or None
    else:
        yomi = None
        base_form = None
        pron = None

    return Token(
        surface=surface,
        postag=postag,
        postag2=postag2,
        postag3=postag3,
        postag4=postag4,
        inflection=inflection,
        conjugation=conjugation,
        base_form=base_form,
        yomi=yomi,
        pron=pron,
    )


class MeCabTokenizer(BaseTokenizer):
    def __init__(
        self,
        user_dictionary_path: Optional[str] = None,
        system_dictionary_path: Optional[str] = None,
        dictionary_format: Optional[str] = None,
    ) -> None:
        from natto import MeCab
        super().__init__(name="mecab")
        options = []
        if isinstance(user_dictionary_path, str):
            options.append("-u {}".format(user_dictionary_path))
        if isinstance(system_dictionary_path, str):
            options.append("-d {}".format(system_dictionary_path))
        self._tokenizer = MeCab(" ".join(options))

        # If dictionary format is not specified,
        # konoha detects it by checking a name of system dictionary.
        # For instance, system_dictionary_path=mecab-ipadic-xxxx -> ipadic and
        #               system_dictionary_path=mecab-unidic-xxxx -> unidic.
        # If system_dictionary_path and dictionary_format are not given,
        # konoha assumes it uses mecab-ipadic (de facto standard).
        # Currently, konoha only supports ipadic. (TODO: unidic)
        if dictionary_format is None:
            if system_dictionary_path is None or "ipadic" in system_dictionary_path.lower():
                self._parse_feature = parse_feature_for_ipadic
            elif "unidic" in system_dictionary_path.lower():
                self._parse_feature = parse_feature_for_unidic
            else:
                raise ValueError(f"Unsupported system dictionary: {system_dictionary_path}")

        else:
            if "ipadic" == dictionary_format.lower():
                self._parse_feature = parse_feature_for_ipadic
            elif "unidic" == dictionary_format.lower():
                self._parse_feature = parse_feature_for_unidic
            else:
                raise ValueError(f"Unsupported dictionary format: {dictionary_format}")

    def tokenize(self, text: str) -> List[Token]:
        return_result = []
        parse_result = self._tokenizer.parse(text).rstrip(" ")
        for elem in parse_result.split("\n")[:-1]:
            return_result.append(self._parse_feature(elem))
        return return_result



import logging
from typing import List
from typing import Optional

from fastapi import APIRouter
from fastapi import HTTPException
from fastapi import Request
from pydantic import BaseModel

from konoha import WordTokenizer


class TokenizeParameter(BaseModel):
    tokenizer: str = "MeCab"
    user_dictionary_path: Optional[str] = None
    system_dictionary_path: Optional[str] = None
    model_path: Optional[str] = None
    mode: Optional[str] = "A"
    dictionary_format: Optional[str] = None
    text: Optional[str] = None
    texts: Optional[List[str]] = None


router = APIRouter()
logger = logging.getLogger(__file__)


def generate_cache_key(params):
    params = params.dict(exclude={"text", "texts"})
    return ".".join(f"{k}-{v}" for k, v in params.items())


@router.post("/api/v1/tokenize")
async def tokenize(params: TokenizeParameter, request: Request):
    if params.texts is not None:
        message = (
            "A parameter `texts` is now unacceptable for /api/v1/tokenize."
            " Please use /api/v1/batch_tokenize instead."
        )
        raise HTTPException(status_code=400, detail=message)

    if params.text is None:
        raise HTTPException(status_code=400, detail="text or texts is required.")

    cache_key = generate_cache_key(params)
    if cache_key in request.app.state.cache:
        logging.info(f"Hit cache: {cache_key}")
        tokenizer = request.app.state.cache[cache_key]
    else:
        logging.info(f"Create tokenizer: {cache_key}")
        try:
            tokenizer = WordTokenizer(
                tokenizer=params.tokenizer,
                user_dictionary_path=params.user_dictionary_path,
                system_dictionary_path=params.system_dictionary_path,
                model_path=params.model_path,
                mode=params.mode,
                dictionary_format=params.dictionary_format,
            )
            request.app.state.cache[cache_key] = tokenizer
        except Exception:
            raise HTTPException(status_code=400, detail="fail to initialize tokenizer")

    tokens = [token.dict() for token in tokenizer.tokenize(params.text)]
    return {"tokens": tokens}



import re
from typing import List
from typing import Match
from typing import Optional
from typing import Pattern


class SentenceTokenizer:

    PERIOD = "。"
    PERIOD_SPECIAL = "__PERIOD__"

    PATTERNS = [
        re.compile(r"（.*?）"),
        re.compile(r"「.*?」"),
    ]

    def __init__(
        self,
        period: Optional[str] = None,
        patterns: Optional[List[Pattern[str]]] = None,
    ) -> None:

        self._period = period or self.PERIOD
        self._patterns = patterns or self.PATTERNS

    def conv_period(self, item: Match) -> str:
        return item.group(0).replace(self._period, SentenceTokenizer.PERIOD_SPECIAL)

    def tokenize(self, document: str) -> List[str]:
        for pattern in self._patterns:
            document = re.sub(pattern, self.conv_period, document)

        result = []
        for line in document.split("\n"):
            line = line.rstrip()
            line = line.replace("\n", "")
            line = line.replace("\r", "")
            line = line.replace(f"{self._period}", f"{self._period}\n")
            sentences = line.split("\n")

            for sentence in sentences:
                if not sentence:
                    continue

                period_special = SentenceTokenizer.PERIOD_SPECIAL
                sentence = sentence.replace(period_special, self._period)
                result.append(sentence)

        return result
